# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"test_fail": "00_test.ipynb",
         "test": "00_test.ipynb",
         "nequals": "00_test.ipynb",
         "test_eq": "00_test.ipynb",
         "test_eq_type": "00_test.ipynb",
         "test_ne": "00_test.ipynb",
         "is_close": "00_test.ipynb",
         "test_close": "00_test.ipynb",
         "test_is": "00_test.ipynb",
         "test_shuffled": "00_test.ipynb",
         "test_stdout": "00_test.ipynb",
         "test_warns": "00_test.ipynb",
         "TEST_IMAGE": "00_test.ipynb",
         "TEST_IMAGE_BW": "00_test.ipynb",
         "test_fig_exists": "00_test.ipynb",
         "defaults": "01_foundation.ipynb",
         "FixSigMeta": "01_foundation.ipynb",
         "PrePostInitMeta": "01_foundation.ipynb",
         "NewChkMeta": "01_foundation.ipynb",
         "BypassNewMeta": "01_foundation.ipynb",
         "copy_func": "01_foundation.ipynb",
         "patch_to": "01_foundation.ipynb",
         "patch": "01_foundation.ipynb",
         "patch_property": "01_foundation.ipynb",
         "use_kwargs_dict": "01_foundation.ipynb",
         "use_kwargs": "01_foundation.ipynb",
         "delegates": "01_foundation.ipynb",
         "method": "01_foundation.ipynb",
         "funcs_kwargs": "01_foundation.ipynb",
         "add_docs": "01_foundation.ipynb",
         "docs": "01_foundation.ipynb",
         "custom_dir": "01_foundation.ipynb",
         "arg0": "01_foundation.ipynb",
         "arg1": "01_foundation.ipynb",
         "arg2": "01_foundation.ipynb",
         "arg3": "01_foundation.ipynb",
         "arg4": "01_foundation.ipynb",
         "bind": "01_foundation.ipynb",
         "GetAttr": "01_foundation.ipynb",
         "delegate_attr": "01_foundation.ipynb",
         "coll_repr": "01_foundation.ipynb",
         "mask2idxs": "01_foundation.ipynb",
         "listable_types": "01_foundation.ipynb",
         "CollBase": "01_foundation.ipynb",
         "cycle": "01_foundation.ipynb",
         "zip_cycle": "01_foundation.ipynb",
         "is_indexer": "01_foundation.ipynb",
         "negate_func": "01_foundation.ipynb",
         "L": "01_foundation.ipynb",
         "ifnone": "02_utils.ipynb",
         "maybe_attr": "02_utils.ipynb",
         "basic_repr": "02_utils.ipynb",
         "get_class": "02_utils.ipynb",
         "mk_class": "02_utils.ipynb",
         "wrap_class": "02_utils.ipynb",
         "ignore_exceptions": "02_utils.ipynb",
         "store_attr": "02_utils.ipynb",
         "attrdict": "02_utils.ipynb",
         "properties": "02_utils.ipynb",
         "camel2snake": "02_utils.ipynb",
         "snake2camel": "02_utils.ipynb",
         "class2attr": "02_utils.ipynb",
         "hasattrs": "02_utils.ipynb",
         "ShowPrint": "02_utils.ipynb",
         "Int": "02_utils.ipynb",
         "Float": "02_utils.ipynb",
         "Str": "02_utils.ipynb",
         "last_index": "02_utils.ipynb",
         "tuplify": "02_utils.ipynb",
         "detuplify": "02_utils.ipynb",
         "replicate": "02_utils.ipynb",
         "uniqueify": "02_utils.ipynb",
         "setify": "02_utils.ipynb",
         "merge": "02_utils.ipynb",
         "is_listy": "02_utils.ipynb",
         "range_of": "02_utils.ipynb",
         "groupby": "02_utils.ipynb",
         "first": "02_utils.ipynb",
         "shufflish": "02_utils.ipynb",
         "IterLen": "02_utils.ipynb",
         "ReindexCollection": "02_utils.ipynb",
         "in_": "02_utils.ipynb",
         "operator.in_": "02_utils.ipynb",
         "Inf": "02_utils.ipynb",
         "true": "02_utils.ipynb",
         "stop": "02_utils.ipynb",
         "gen": "02_utils.ipynb",
         "chunked": "02_utils.ipynb",
         "num_methods": "02_utils.ipynb",
         "rnum_methods": "02_utils.ipynb",
         "inum_methods": "02_utils.ipynb",
         "Tuple": "02_utils.ipynb",
         "trace": "02_utils.ipynb",
         "compose": "02_utils.ipynb",
         "maps": "02_utils.ipynb",
         "partialler": "02_utils.ipynb",
         "mapped": "02_utils.ipynb",
         "instantiate": "02_utils.ipynb",
         "using_attr": "02_utils.ipynb",
         "log_args": "02_utils.ipynb",
         "Self": "02_utils.ipynb",
         "Path.readlines": "02_utils.ipynb",
         "Path.read": "02_utils.ipynb",
         "Path.write": "02_utils.ipynb",
         "Path.save": "02_utils.ipynb",
         "Path.load": "02_utils.ipynb",
         "Path.ls": "02_utils.ipynb",
         "Path.__repr__": "02_utils.ipynb",
         "bunzip": "02_utils.ipynb",
         "join_path_file": "02_utils.ipynb",
         "remove_patches_path": "02_utils.ipynb",
         "sort_by_run": "02_utils.ipynb",
         "PrettyString": "02_utils.ipynb",
         "round_multiple": "02_utils.ipynb",
         "even_mults": "02_utils.ipynb",
         "num_cpus": "02_utils.ipynb",
         "defaults.cpus": "02_utils.ipynb",
         "add_props": "02_utils.ipynb",
         "change_attr": "02_utils.ipynb",
         "change_attrs": "02_utils.ipynb",
         "ContextManagers": "02_utils.ipynb",
         "set_num_threads": "02_utils.ipynb",
         "ProcessPoolExecutor": "02_utils.ipynb",
         "parallel": "02_utils.ipynb",
         "parallel_chunks": "02_utils.ipynb",
         "run_procs": "02_utils.ipynb",
         "parallel_gen": "02_utils.ipynb",
         "in_ipython": "02_utils.ipynb",
         "in_colab": "02_utils.ipynb",
         "in_notebook": "02_utils.ipynb",
         "type_hints": "03_dispatch.ipynb",
         "anno_ret": "03_dispatch.ipynb",
         "cmp_instance": "03_dispatch.ipynb",
         "TypeDispatch": "03_dispatch.ipynb",
         "DispatchReg": "03_dispatch.ipynb",
         "typedispatch": "03_dispatch.ipynb",
         "retain_meta": "03_dispatch.ipynb",
         "default_set_meta": "03_dispatch.ipynb",
         "retain_type": "03_dispatch.ipynb",
         "retain_types": "03_dispatch.ipynb",
         "explode_types": "03_dispatch.ipynb",
         "Transform": "04_transform.ipynb",
         "InplaceTransform": "04_transform.ipynb",
         "DisplayedTransform": "04_transform.ipynb",
         "ItemTransform": "04_transform.ipynb",
         "get_func": "04_transform.ipynb",
         "Func": "04_transform.ipynb",
         "Sig": "04_transform.ipynb",
         "compose_tfms": "04_transform.ipynb",
         "mk_transform": "04_transform.ipynb",
         "gather_attrs": "04_transform.ipynb",
         "gather_attr_names": "04_transform.ipynb",
         "Pipeline": "04_transform.ipynb"}

modules = ["test.py",
           "foundation.py",
           "utils.py",
           "dispatch.py",
           "transform.py"]

doc_url = "https://fastcore.fast.ai/"

git_url = "https://github.com/fastai/fastcore/tree/master/"

def custom_doc_links(name): return None
